Line[] lines;
PVector centerPos;
float step = 0;
float drawLength = 1000;
float radius = 0;
float lastMillis = 0;
float maxRCenter;
float maxRAmp;
float minXCenter;
float maxXCenter;
float maxXAmp;
float minYCenter;
float maxYCenter;
float maxYAmp;
float maxRadiusCenter;
float maxRadiusAmp;
boolean saved = false;
boolean saveThis = false;
double drawArea = 0;

//-------------------------------------------------------------------------------------------------//
//----------------------------FEEL FREE TO EDIT THIS STUFF-----------------------------------------//
//-------------------------------------------------------------------------------------------------//

//Controls:
//  - Shift + S: Save current harmonograph (on fade out)
//  - Shift + C: Fade out instantly
//  - C: Disable automatic fade out for this harmonograph (turns back on for the next one)
//Also note: if shuffle is turned on, many of these won't do anything (since shuffle will overrite them)
//Values that are overwritten are marked with an asterisk in the comment

//Resolution
int resolutionWidth = 1920;       //Horizontal window size
int resolutionHeight = 1080;      //Vertical window size

//Overall program speed values
int maxFrameRate = 60;            //the maximum framerate - turn way down to see exactly what's happening each frame
float stepRate = 1;               //defines the overall speed of everything.
int stepsPerFrame = 1;            //multiplies the speed directly. Murders performance, but doesn't change the result at all

//Line behaviour values
int noLines = 12;                 //* how many lines there are
float lengthFactor = 1;           //line length = lengthFactor * current radius (note that additional lines are drawn past this length)
float angleOffset = PI / 1.2;     //* offset of line angle (PI / 2.0 makes them tangential)

//Cosmetic values
float lineWidth = 1.0;            //line width in px (best to keep at 1)
float lineColStart = 0.0;         //* The hue value (0 to 1) that the first line has
float lineColSpread = 1.0;        //* The spread of hues that is distributed over the ring
float saturation = 1.0;           //line saturation (0 to 1)
float brightness = 1.0;           //line brightness (0 to 1)
float alpha = 0.03;                //line opacity (0.008 to 1)
float bgBrightness = 0.0;         //background brightness (0 to 1)

//Harmonograph values
float rotRate = 0.01;             //* speed of lines rotating around ring

float radiusCenter = 100;         //* middle point of radius oscillation
float radiusAmp = 500;            //* variance of radius oscillation
float radiusFreq = 0.005;         //* frequency of radius oscillation

boolean usePolar = false;         //* If this is false, uses X and Y (cartesian coordinates)
                                  //* If it's true, uses R and T (polar coordinates)
                                  
float xCenter = 500;              //* middle point of X-position oscillation
float xAmp = 0;                   //* variance of X-position oscillation
float xFreq = 0.001;              //* frequency of X-position oscillation

float yCenter = 500;              //* middle point of Y-position oscillation
float yAmp = 0;                   //* variance of Y-position oscillation
float yFreq = 0.003;              //* frequency of Y-position oscillation

float rCenter = 50;               //* middle point of r-position oscillation
float rAmp = 0;                   //* variance of r-position oscillation
float rFreq = 0.001;              //* frequency of r-position oscillation

float tCenter = 0;                //* middle point of t-position oscillation
float tAmp = 0;                   //* variance of t-position oscillation
float tFreq = 0.001;              //* frequency of t-position oscillation

//Noise values
boolean usePerlin = false;
int noOctaves = 4;
float lacunarity = 2.0;
float persistence = 0.5;
float perlinFreqMod = 2;
float perlinShift = 1;

//Shuffle behaviours
boolean shuffle = true;           //toggles period shuffling on or off
boolean saveOnShuffle = false;    //if on, every shuffle will save a screengrab
boolean logShuffle = false;       //if on, logs all shuffle value changes in the console
float minShuffleTimer = 9.0;
float maxShuffleTimer = 18.0;
float shuffleTimer = 9.0;         //shuffle period in seconds
float fadeTime = 1.0;
float shuffleProb = 1.0;          //the probability (0 to 1) that each parameter will be randomised on a shuffle
float zeroProb = 0.3;             //the probability that a value will be reset to default on a randomisation

//Shuffle constraints
int maxNoLines = 24;              //Maximum number of lines around ring (minimum 3)
float maxRotRate = 0.01;          //Maximum line rotation speed
float maxFreq = 0.02;             //Maximum frequency for all periodic functions
float maxTAmp = 2 * PI;           //Maximum amplitude for ring t oscillation
float xyMargin = 0.1;             //Fraction of window size defining the margins for xy oscillation
float xyAmpM = 0.5;               //Fraction of window size defining the max amplitude for xy oscillation
float rMargin = 0.5;              //Fraction of window size defining the margins for r oscillation
float rAmpM = 0.33;               //Fraction of window size defining the max amplitude for r oscillation
float radMargin = 0.25;           //Fraction of window size defining the margins for radius oscillation
float radAmpM = 0.25;             //Fraction of window size defining the max amplitude for radius oscillation

PVector mousePos;
boolean useMouse = false;

//-------------------------------------------------------------------------------------------------//
//-----------------------UNDER HERE IS THE CODE. EDIT AT OWN RISK!---------------------------------//
//-------------------------------------------------------------------------------------------------//

void setup() {
  resolutionWidth = displayWidth;
  resolutionHeight = displayHeight;
  frame.setTitle("Line Circles");
  
  //set up the window size, colorMode and maximum frameRate
  size(resolutionWidth,resolutionHeight);
  colorMode(HSB, 1.0);
  //frameRate(maxFrameRate);
  frameRate(10000);
  
  //set certain values that are dependent on window size
  centerPos = new PVector(width/2, height/2);
  mousePos = new PVector(width/2, height/2);
  minXCenter = ((width + height) / 2.0) * xyMargin;
  maxXCenter = width - (((width + height) / 2.0) * xyMargin);
  maxXAmp = width * xyAmpM;
  minYCenter = ((width + height) / 2.0) * xyMargin;
  maxYCenter = height - (((width + height) / 2.0) * xyMargin);
  maxYAmp = height * xyAmpM;
  maxRCenter = ((width + height) / 2.0) * rMargin;
  maxRAmp = ((width + height) / 2.0) * rAmpM;
  maxRadiusCenter = ((width + height) / 2.0) * radMargin;
  maxRadiusAmp = ((width + height) / 2.0) * radAmpM;

  //finally, we do our first shuffle
  Shuffle();
  
  background(0.0, 0.0, bgBrightness);
  
  noCursor();
} 

float getPerlin(float min, float amp, float pos, int octaves, float lac, float pers) {
  float sum = min + amp * noise(min + amp, pos);
  float freq = 1.;
  float ampl = 1.;
  float range = 1.;
  for(int o = 1; o < octaves; o++) {
     freq *= lac;
     amp *= pers;
     range += ampl;
     sum += noise(min + amp, pos * freq) * ampl * amp;
  }
  return ((sum/range) + perlinShift);
}

void draw() {
  //housekeeping...
  colorMode(HSB, 1.0);
  //frame.setTitle(nf(frameRate, 2, 1));
  //frame.setTitle("DrawArea: " + drawArea);
  
  //background(0);
  //update radius, centerPos and drawLength
  if(!usePerlin) radius = radiusCenter + (radiusAmp * sin(step * radiusFreq));
  else radius = getPerlin(radiusCenter, radiusAmp, step * radiusFreq * perlinFreqMod, noOctaves, lacunarity, persistence);
  if(usePolar) {
    float r; 
    if(!usePerlin) r = rCenter + (rAmp * sin(step * rFreq));
    else r = getPerlin(rCenter, rAmp, step * rFreq * perlinFreqMod, noOctaves, lacunarity, persistence);
    float t; 
    if(!usePerlin) t = tCenter + (tAmp * sin(step * tFreq));
    else t = getPerlin(tCenter, tAmp, step * tFreq * perlinFreqMod, noOctaves, lacunarity, persistence);
    
    //convert polar to cartesian
    centerPos.x = (width / 2.0) + (r * cos(t));
    centerPos.y = (height / 2.0) + (r * sin(t));
  } else {
    if(!usePerlin) centerPos.x = xCenter + (xAmp * sin(step * xFreq));
    else centerPos.x = getPerlin(xCenter, xAmp, step * xFreq * perlinFreqMod, noOctaves, lacunarity, persistence);
    if(!usePerlin) centerPos.y = yCenter + (yAmp * sin(step * yFreq));
    else centerPos.y = getPerlin(yCenter, yAmp, step * yFreq * perlinFreqMod, noOctaves, lacunarity, persistence);
  }
  drawLength = radius * lengthFactor;
  
  float lerpSpeed = map(radius, radiusCenter - radiusAmp, radiusCenter + radiusAmp, 0.0, 1.0);
  //print("LerpSpeed: " + lerpSpeed);
  lerpSpeed = 0.01 * pow(10, -3.0 * lerpSpeed + 3.0);
  lerpSpeed = max(0.0, min(1.0, lerpSpeed));
  //println(" to " + lerpSpeed);
  mousePos.x = lerp(mousePos.x, mouseX, lerpSpeed);
  mousePos.y = lerp(mousePos.y, mouseY, lerpSpeed);
  
  if(useMouse) {
    centerPos.x = mousePos.x;
    centerPos.y = mousePos.y;
  }
  
  for(int i = 0; i < lines.length; i++) {
    float theta = lines[i].drawAngle;
    if(!usePerlin) theta += rotRate;//(((float)(mouseX - (width/2))/(float)(width/2)))/10.0;//theta += rotRate;
    else theta += rotRate * getPerlin(0.0, 1.0, step * perlinFreqMod, noOctaves, lacunarity, persistence);
    if(theta > 2 * PI) { theta -= 2 * PI; }
    
    //update line angle and start position
    lines[i].startPos.set(centerPos.x + radius * cos(theta + angleOffset), centerPos.y + radius * sin(theta + angleOffset));
    lines[i].drawAngle = theta;
    
    //draw the line
    drawArea += lines[i].Draw();
  }

  //we update t each time we draw anything
  step += stepRate;
  
  if((millis() - lastMillis) / 1000.0 > shuffleTimer && shuffle) {
    lastMillis = millis();
    Shuffle(); 
    shuffleTimer = random(maxShuffleTimer - minShuffleTimer) + minShuffleTimer;
    saved = false;
  } else if((millis() - lastMillis) / 1000.0 > (shuffleTimer - fadeTime) && shuffle) {
     noStroke();
     fill(0, 0, 0, (((millis() - lastMillis) / 1000.0) - (shuffleTimer - fadeTime)) / fadeTime);
     rect(0, 0, width, height);
     
     //save screenshot
     if((saveOnShuffle || saveThis) && !saved) {
       String path = "/Screenshots/";
       path += "HarmonographGrab_" + year() + "-" + month() + "-" + day() + "_" + hour() + "-" + minute() + "-" + second() + ".png";
       save(path);
       println("Saved as: " + path);
       saved = true;
       if(saveThis) saveThis = false;
     }
  }
}

void Shuffle() {
   if(logShuffle) { println(""); }
   if(logShuffle) { println("-----------Shuffle-----------"); }
  
   //randomise no lines
   if(random(1) < shuffleProb) {
     noLines = 3 + (int)random(maxNoLines - 3);
     if(logShuffle) { println("noLines: " + noLines); }
   }
   
   //randomise line colour start
   if(random(1) < shuffleProb) {
      lineColStart = random(1);
      if(logShuffle) { println("lineColStart: " + lineColStart); }
   }
   
   //randomise line colour spread
   if(random(1) < shuffleProb) {
     lineColSpread = random(1); 
     if(logShuffle) { println("lineColSpread: " + lineColSpread); }
   }
   
   //reset lines
   lines = new Line[noLines];
   for(int i = 0; i < lines.length; i++) {
    float iF = (float)i / (float)noLines;
    
    //each line is distributed around the ring
    float theta = iF * 2 * PI;
    lines[i] = new Line(
      new PVector(centerPos.x + radius * cos(theta), centerPos.y + radius * sin(theta)),
      theta - angleOffset,
      color((lineColStart + (lineColSpread * iF)) % 1.0, saturation, brightness, alpha), //rainbow around the ring
      lineWidth
    );
   }
   
   //randomise angle offset
   if(random(1) < shuffleProb) {
     angleOffset = random(2 * PI); 
     if(logShuffle) { println("angleOffset: " + angleOffset); }
   }
   
   //randomise rotRate
   if(random(1) < shuffleProb) {
     rotRate = random(maxRotRate); 
     if(logShuffle) { println("rotRate: " + rotRate); }
   }
   
   //randomise usePolar
   if(random(1) < shuffleProb) {
     usePolar = !usePolar; 
     if(logShuffle) { println("usePolar: " + usePolar); }
   }
   
   if(usePolar) {
     //randomise rCenter
     if(random(1) < shuffleProb) {
       rCenter = random(maxRCenter); 
       if(logShuffle) { println("rCenter: " + rCenter); }
     }
     
     //randomise rAmp
     if(random(1) < shuffleProb) {
       if(random(1) > zeroProb) {
         rAmp = random(maxRAmp);
       } else { rAmp = 0; }
       if(logShuffle) { println("rAmp: " + rAmp); }
     }
     
     //randomise rFreq
     if(random(1) < shuffleProb) {
         rFreq = random(maxFreq); 
         if(logShuffle) { println("rFreq: " + rFreq); }
     }
     
     //randomise tCenter
     if(random(1) < shuffleProb) {
       tCenter = random(2 * PI); 
       if(logShuffle) { println("tCenter: " + tCenter); }
     }
     
     //randomise tAmp
     if(random(1) < shuffleProb) {
       if(random(1) > zeroProb) {
         tAmp = random(maxTAmp);
       } else { tAmp = 0; }
       if(logShuffle) { println("tAmp: " + tAmp); }
     }
     
     //randomise tFreq
     if(random(1) < shuffleProb) {
         tFreq = random(maxFreq); 
         if(logShuffle) { println("tFreq: " + tFreq); }
     }
   } else {
     //randomise xCenter
     if(random(1) < shuffleProb) {
       if(random(1) > zeroProb) {
         xCenter = minXCenter + random(maxXCenter - minXCenter); 
       } else { xCenter = width/2.0; }
       if(logShuffle) { println("xCenter: " + xCenter); }
     }
     
     //randomise xAmp
     if(random(1) < shuffleProb) {
       if(random(1) > zeroProb) {
         xAmp = random(maxXAmp);
       } else { xAmp = 0; }
       if(logShuffle) { println("xAmp: " + xAmp); }
     }
     
     //randomise xFreq
     if(random(1) < shuffleProb) {
         xFreq = random(maxFreq); 
         if(logShuffle) { println("xFreq: " + xFreq); }
     }
     
     //randomise yCenter
     if(random(1) < shuffleProb) {
       if(random(1) > zeroProb) {
         yCenter = minYCenter + random(maxYCenter - minYCenter); 
       } else { yCenter = height/2.0; }
       if(logShuffle) { println("yCenter: " + yCenter); }
     }
     
     //randomise yAmp
     if(random(1) < shuffleProb) {
       if(random(1) > zeroProb) {
         yAmp = random(maxYAmp);
       } else { yAmp = 0; }
       if(logShuffle) { println("yAmp: " + yAmp); }
     }
     
     //randomise yFreq
     if(random(1) < shuffleProb) {
         yFreq = random(maxFreq);
         if(logShuffle) { println("yFreq: " + yFreq); }
     }
   }
   
   //randomise radiusCenter
     if(random(1) < shuffleProb) {
       radiusCenter = random(maxRadiusCenter);
       if(logShuffle) { println("radiusCenter: " + radiusCenter); }
     }
     
     //randomise radiusAmp
     if(random(1) < shuffleProb) {
       if(random(1) > zeroProb) {
         radiusAmp = random(maxRadiusAmp);
       } else { radiusAmp = 0; }
       if(logShuffle) { println("radiusAmp: " + radiusAmp); }
     }
     
     //randomise radiusFreq
     if(random(1) < shuffleProb) {
         radiusFreq = random(maxFreq); 
         if(logShuffle) { println("radiusFreq: " + radiusFreq); }
     }
     
     //reset frame
     background(0, 0, bgBrightness);
     println("drawArea: " + drawArea + " million pixels");
     drawArea = 0;
}

void keyPressed() {
  if(key == 'S') {
    saveThis = true;
  } else if(key == 'C') {
    lastMillis = millis() + fadeTime*1000.0 - shuffleTimer*1000.0; 
  } else if(key == 'c') {
    lastMillis = millis() + 1000000.0; 
  } else if(key == 'm') { useMouse = !useMouse; }
}
